import { withLatestFrom, map, filter, takeWhile, tap, finalize } from 'rxjs/operators';
import { isObservable, of, Observable } from 'rxjs';
import { isIterable } from 'rxjs/internal-compatibility';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __read(o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
}

function __spread() {
    for (var ar = [], i = 0; i < arguments.length; i++)
        ar = ar.concat(__read(arguments[i]));
    return ar;
}

/**
 * @packageDocumentation
 * @module String
 */
/**
 * @private
 * @internal
 * @param input
 */

function createOrReturnObservable(input) {
  return isObservable(input) ? input : of(input);
}

/**
 * Returns an Observable that emits an array of strings. The string is returned from String.charAt using the passed
 * index or array of indexes.  The output is always an Array containing the string, or empty string if nothing found.
 *
 * @category Query
 *
 * @param positions Single or list of index values to return the character at
 *
 * @example
 * Return the character at index `1`
 * ```ts
 * of('RxJS Ninja').pipe(charAt(1)).subscribe();
 * ```
 * Output: `['x']`
 *
 * @example
 * Return the character at index `1` and `7`
 * ```ts
 * of('RxJS Ninja').pipe(charAt([1, 7])).subscribe();
 * ```
 * Output: `['x', 'n']`
 *
 * @returns Observable that emits an Array of string values
 */

function charAt(positions) {
  var position$ = createOrReturnObservable(positions);
  return function (source) {
    return source.pipe(withLatestFrom(position$), map(function (_a) {
      var _b = __read(_a, 2),
          value = _b[0],
          positionsValue = _b[1];

      return typeof positionsValue === 'number' ? [value.charAt(positionsValue)] : __spread(positionsValue).map(function (v) {
        return value.charAt(v);
      });
    }));
  };
}

/**
 * Returns an Observable that emits an array of numbers. The number is returned from String.charCodeAt using the passed
 * index or array of indexes.  The output is always an Array containing the number, or NaN.
 *
 * @category Query
 *
 * @param positions Single or list of index values to return the character at
 *
 * @example
 * Return the character code of the character at index `1`
 * ```ts
 * of('RxJS Ninja').pipe(charCodeAt(1)).subscribe();
 * ```
 * Output: `120`
 *
 * @returns Observable that emits an Array of number values
 */

function charCodeAt(positions) {
  var position$ = createOrReturnObservable(positions);
  return function (source) {
    return source.pipe(withLatestFrom(position$), map(function (_a) {
      var _b = __read(_a, 2),
          value = _b[0],
          inputValue = _b[1];

      return typeof inputValue === 'number' ? [value.charCodeAt(inputValue)] : __spread(inputValue).map(function (v) {
        return value.charCodeAt(v);
      });
    }));
  };
}

/**
 * Returns an Observable that emits an array of numbers. The number is returned from String.codePointAt using the passed
 * index or array of indexes.  The output is always an Array containing the number, or NaN.
 *
 * @category Query
 *
 * @remarks By default String.codePointAt returns undefined, this operator returns NaN instead matching [[charCodeAt]]
 *
 * @param positions Single or list of index values to return the character at
 *
 * @example
 * Return the code point of the character at index `1`
 * ```ts
 * of('☃★♲').pipe(codePointAt(1)).subscribe();
 * ```
 * Output: `9733`
 *
 * @returns Observable that emits a number that is a code point
 */

function codePointAt(positions) {
  var position$ = createOrReturnObservable(positions);
  return function (source) {
    return source.pipe(withLatestFrom(position$), map(function (_a) {
      var _b;

      var _c = __read(_a, 2),
          value = _c[0],
          inputValue = _c[1];

      return typeof inputValue === 'number' ? [(_b = value.codePointAt(inputValue)) !== null && _b !== void 0 ? _b : NaN] : __spread(inputValue).map(function (v) {
        var _a;

        return (_a = value.codePointAt(v)) !== null && _a !== void 0 ? _a : NaN;
      });
    }));
  };
}

/**
 * Returns an Observable that emits a string that is the source string concatenated with the passed input to the
 * operator using Sting.concat
 *
 * @category Modify
 *
 * @param input Single or list of arguments to concatenate with the source string
 *
 * @example
 * Return a string that is a source appended with a string
 * ```ts
 * of('RxJS').pipe(concat('Ninja')).subscribe();
 * ```
 * Output: `RxJSNinja`
 *
 * @example
 * Return a string that is a source appended with a array of strings
 * ```ts
 * of('RxJS').pipe(concat([' ', 'Ninja'])).subscribe();
 * ```
 * Output: `RxJS Ninja`
 *
 * @example
 * Return a string that is a source appended with an Observable strings
 * ```ts
 * of('RxJS').pipe(concat(of([' ', 'Ninja']))).subscribe();
 * ```
 * Output: `RxJS Ninja`
 *
 * @returns Observable that emits a string
 */

function concat(input) {
  var input$ = createOrReturnObservable(input);
  return function (source) {
    return source.pipe(withLatestFrom(input$), map(function (_a) {
      var _b = __read(_a, 2),
          value = _b[0],
          inputValue = _b[1];

      return typeof inputValue === 'string' ? value.concat(inputValue) : value.concat.apply(value, __spread(inputValue));
    }));
  };
}

/**
 * Returns an Observable that emits a boolean value where the source string ends with the passed string parameter using
 * String.endsWith
 *
 * @category Query
 *
 * @see The [[filterEndsWith]] operator returns the string value
 *
 * @param search The string to check the source ends with
 * @param maxLength Optional length of the string to check
 *
 * @example
 * Return if the source string ends with the `S` character
 * ```ts
 * from(['RxJS', 'Ninja', 'Tests']).pipe(endsWith('S')).subscribe();
 * ```
 * Output: `true, false, false`
 *
 * @example
 * Return if the source string at up to length `4` ends with the `t` character
 * ```ts
 * from(['RxJS', 'Ninja', 'Tests']).pipe(endsWith('t', 4)).subscribe();
 * ```
 * Output: `false, false, true`
 *
 * @returns Observable that emits a boolean of the source string ending with the passed input
 */

function endsWith(search, maxLength) {
  var search$ = createOrReturnObservable(search);
  var maxLength$ = createOrReturnObservable(maxLength);
  return function (source) {
    return source.pipe(withLatestFrom(search$, maxLength$), map(function (_a) {
      var _b = __read(_a, 3),
          value = _b[0],
          searchValue = _b[1],
          maxLengthValue = _b[2];

      return value.endsWith(searchValue, maxLengthValue);
    }));
  };
}

/**
 * Returns an Observable that emits a string where the source string ends with the passed ending using String.endsWith
 *
 * @category Filter
 *
 * @see The [[endsWith]] operator returns the boolean value
 *
 * @param search The string to check the source ends with
 * @param maxLength Optional length of the string to check
 *
 * @example
 * Return a string where the source string ends with the `S` character
 * ```ts
 * from(['RxJS', 'Ninja', 'Tests']).pipe(filterEndsWith('S')).subscribe();
 * ```
 * Output: `RxJS`
 *
 * @example
 * Return a string where the source string at up to length `4` ends with the `t` character
 * ```ts
 * from(['RxJS', 'Ninja', 'Tests']).pipe(filterEndsWith('t', 4)).subscribe();
 * ```
 * Output: `Tests`
 *
 * @returns Observable that emits a string
 */

function filterEndsWith(search, maxLength) {
  var search$ = createOrReturnObservable(search);
  var maxLength$ = createOrReturnObservable(maxLength);
  return function (source) {
    return source.pipe(withLatestFrom(search$, maxLength$), map(function (_a) {
      var _b = __read(_a, 3),
          value = _b[0],
          searchValue = _b[1],
          maxLengthValue = _b[2];

      return value.endsWith(searchValue, maxLengthValue) ? value : '';
    }), filter(function (val) {
      return Boolean(val);
    }));
  };
}

/**
 * Returns an Observable that emits a string where the source string contains with the passed search string using
 * String.includes
 *
 * @category Filter
 *
 * @see The [[includes]] operator returns the boolean value
 *
 * @param search The string to check the source ends with
 *
 * @example
 * Return a string where the source string includes 'JS'
 * ```ts
 * from(['RxJS', 'Ninja', 'Tests']).pipe(filterIncludes('JS')).subscribe();
 * ```
 * Output: `RxJS`
 *
 * @returns Observable that emits a string
 */

function filterIncludes(search) {
  var search$ = createOrReturnObservable(search);
  return function (source) {
    return source.pipe(withLatestFrom(search$), map(function (_a) {
      var _b = __read(_a, 2),
          value = _b[0],
          searchValue = _b[1];

      return value.includes(searchValue) ? value : '';
    }), filter(function (value) {
      return Boolean(value);
    }));
  };
}

/**
 * Returns an Observable that emits a string where the source string starts with the passed string using
 * String.startsWith
 *
 * @category Filter
 *
 * @see The [[startsWith]] operator returns the boolean value
 *
 * @param search The string to check the source starts with
 * @param startIndex Optional index to start the check from
 *
 * @example
 * Return a string where the source string starts with the `N` character
 * ```ts
 * from(['RxJS', 'Ninja', 'Tests']).pipe(filterStartsWith('N')).subscribe();
 * ```
 * Output: `Ninja`
 *
 * @example
 * Return a string where the source string starts with `t` character from index `3`
 * ```ts
 * from(['RxJS', 'Ninja', 'Tests']).pipe(filterStartsWith('t', 3)).subscribe();
 * ```
 * Output: `Tests`
 *
 * @returns Observable that emits a string
 */

function filterStartsWith(search, startIndex) {
  var search$ = createOrReturnObservable(search);
  var startIndex$ = createOrReturnObservable(startIndex);
  return function (source) {
    return source.pipe(withLatestFrom(search$, startIndex$), map(function (_a) {
      var _b = __read(_a, 3),
          value = _b[0],
          searchValue = _b[1],
          startIndexValue = _b[2];

      return value.startsWith(searchValue, startIndexValue) ? value : '';
    }), filter(function (value) {
      return Boolean(value);
    }));
  };
}

/**
 * Returns an Observable that emits a string made from character codes using String.fromCharCode
 *
 * @remarks This operator will emit a single string for all input passed including arrays
 *
 * @see The [[mapCharCode]] operator can be used to map an Observable source of char codes to strings
 *
 * @category Create
 *
 * @param input Single or list of character codes to convert to a string
 *
 * @example
 * Return a string from character code array
 * ```ts
 * fromCharCode([82, 120, 74, 83]).subscribe();
 * ```
 * Output: `RxJS`
 *
 * @example
 * Return a string from character code Observable
 * ```ts
 * fromCharCode(of([82, 120, 74, 83]).subscribe();
 * ```
 * Output: `RxJS`
 *
 * @returns Observable that emits a string
 */

function fromCharCode(input) {
  return createOrReturnObservable(input).pipe(map(function (value) {
    return isIterable(value) ? __spread(value) : [value];
  }), map(function (value) {
    return String.fromCharCode.apply(String, __spread(value));
  }));
}

/**
 * Returns an Observable that emits a string made from code points using String.fromCodePoint
 *
 * @remarks This operator will emit a single string for all input passed including arrays
 *
 * @category Create
 *
 * @param input Single or list of code points to convert to a string
 *
 * @example
 * Return a string from code points array
 * ```ts
 * fromCodePoint([9731, 9733, 9842]).subscribe();
 * ```
 * Output: `☃★♲`
 *
 * @example
 * Return a string from code points Observable
 * ```ts
 * fromCodePoint(of([9731, 9733, 9842]).subscribe();
 * ```
 * Output: `☃★♲`
 *
 * @returns Observable that emits a string
 */

function fromCodePoint(input) {
  return createOrReturnObservable(input).pipe(map(function (value) {
    return typeof value === 'number' ? [value] : __spread(value);
  }), map(function (value) {
    return String.fromCodePoint.apply(String, __spread(value));
  }));
}

/**
 * Returns an Observable that emits strings from any an argument list of strings or  supported Observable, Promise or
 * Array-like source
 *
 * @category Create
 *
 * @remarks This operator is a type-safe {@link https://rxjs.dev/api/index/function/from|from} and will emit only
 * strings, also unlike `from` a single string is not converted into an array-like.
 *
 * @param input Argument list, Observable input, Promise or Array of strings
 *
 * @example
 * Return a reversed string from an array list of strings
 * ```ts
 * fromString(['RxJS', 'Ninja']).pipe(reverse()).subscribe();
 * ```
 * Output: `'SJxR', 'ajniN'`
 *
 * @example
 * Return a reversed string from an Observable array list of strings
 * ```ts
 * fromString(of(['RxJS', 'Ninja'])).pipe(reverse()).subscribe();
 * ```
 * Output: `'SJxR', 'ajniN'`
 *
 * @returns Observable that emits a string
 */

function fromString(input) {
  return new Observable(function (subscriber) {
    createOrReturnObservable(input).pipe(takeWhile(function () {
      return !subscriber.closed;
    }), map(function (value) {
      return typeof value === 'string' ? [value] : __spread(value);
    }), tap(function (value) {
      for (var i = 0; i < value.length; i++) {
        subscriber.next(value[i]);
      }
    }), finalize(function () {
      return !subscriber.closed && subscriber.complete();
    })).subscribe();
    /* istanbul ignore next-line */

    return function () {
      return !subscriber.closed && subscriber.complete();
    };
  });
}

/**
 * Returns an Observable that emits a string made from a source unicode string using String.normalize
 *
 * @category Create
 *
 * @remarks This operator is a type-safe {@link https://rxjs.dev/api/index/function/from|from} and will emit only
 * strings, also unlike `from` a single string is not converted into an array-like.
 *
 * @param input Single or list of Unicode character to convert to a string
 * @param form The Unicode Normalization Form to decode the string with
 *
 * @example
 * Returns a string from Unicode characters
 * ```ts
 * fromUnicode('\u0041\u006d\u00e9\u006c\u0069\u0065').subscribe();
 * ```
 * Output: `Amélie`
 *
 * @returns Observable that emits a string
 */

function fromUnicode(input, form) {
  return new Observable(function (subscriber) {
    createOrReturnObservable(input).pipe(takeWhile(function () {
      return !subscriber.closed;
    }), map(function (value) {
      return typeof value === 'string' ? [value] : __spread(value);
    }), tap(function (value) {
      for (var i = 0; i < value.length; i++) {
        subscriber.next(value[i].normalize(form));
      }
    }), finalize(function () {
      return !subscriber.closed && subscriber.complete();
    })).subscribe();
    /* istanbul ignore next-line */

    return function () {
      return !subscriber.closed && subscriber.complete();
    };
  });
}

/**
 * Returns an Observable that emits a boolean where the source string contains with the passed search string using
 * String.includes
 *
 * @category Query
 *
 * @see The [[filterIncludes]] operator returns the string value
 *
 * @param search The string to check the source ends with
 *
 * @example
 * Return a boolean where the source string includes 'JS'
 * ```ts
 * from(['RxJS', 'Ninja', 'Tests']).pipe(includes('JS')).subscribe();
 * ```
 * Output: `true, false, false`
 *
 * @returns Observable that emits a boolean
 */

function includes(search) {
  var search$ = createOrReturnObservable(search);
  return function (source) {
    return source.pipe(withLatestFrom(search$), map(function (_a) {
      var _b = __read(_a, 2),
          value = _b[0],
          searchValue = _b[1];

      return value.includes(searchValue);
    }));
  };
}

/**
 * Returns an Observable that emits a number of the first index from the source string where the search string begins
 * using String.indexOf
 *
 * @category Query
 *
 * @param search The string to search in the source string
 * @param startIndex Optional start position if not from the beginning of the string
 *
 * @example
 * Returns the first index of `RxJS` in the string
 * ```ts
 * of('Hello RxJS Ninja, RxJS Rocks').pipe(indexOf('RxJS')).subscribe();
 * ```
 * Output: `6`
 *
 * @example
 * Returns the first index of `RxJS` in the string, starting from index `8`
 * ```ts
 * of('Hello RxJS Ninja, RxJS Rocks').pipe(indexOf('RxJS', 8)).subscribe();
 * ```
 * Output: `18`
 *
 * @returns Observable that emits a number that is the first index of the search string in the source string
 */

function indexOf(search, startIndex) {
  var search$ = createOrReturnObservable(search);
  var startIndex$ = createOrReturnObservable(startIndex);
  return function (source) {
    return source.pipe(withLatestFrom(search$, startIndex$), map(function (_a) {
      var _b = __read(_a, 3),
          value = _b[0],
          searchValue = _b[1],
          startIndexValue = _b[2];

      return value.indexOf(searchValue, startIndexValue);
    }));
  };
}

/**
 * Returns an Observable that emits a string value from an Array or `Set` of values joined to a single string using
 * the separator.
 *
 * @category Convert
 *
 * @typeParam T The type of value in the source Array or `Map`
 *
 * @param separator The character to join text with, default is a space character
 *
 * @returns Observable that emits a string from an Array or`Set of values
 */

function join(separator) {
  if (separator === void 0) {
    separator = ' ';
  }

  var separator$ = createOrReturnObservable(separator);
  return function (source) {
    return source.pipe(withLatestFrom(separator$), map(function (_a) {
      var _b = __read(_a, 2),
          value = _b[0],
          separatorValue = _b[1];

      return __spread(value).join(separatorValue);
    }));
  };
}

/**
 * Returns an Observable that emits a number of the last index from the source string where the search string begins
 * using String.lastIndexOf
 *
 * @category Query
 *
 * @param search The string to search in the source string
 * @param lastIndex The index of the last character in the string to search up to
 *
 * @example
 * Returns the first index of `RxJS` in the string
 * ```ts
 * of('Hello RxJS Ninja, RxJS Rocks').pipe(lastIndexOf('RxJS')).subscribe();
 * ```
 * Output: `18`
 *
 * @example
 * Returns the first index of `RxJS` in the string, ending at index `15`
 * ```ts
 * of('Hello RxJS Ninja, RxJS Rocks').pipe(lastIndexOf('RxJS', 15)).subscribe();
 * ```
 * Output: `6`
 *
 * @returns Observable that emits a number that is the last index of the search string in the source string
 */

function lastIndexOf(search, lastIndex) {
  var search$ = createOrReturnObservable(search);
  var lastIndex$ = createOrReturnObservable(lastIndex);
  return function (source) {
    return source.pipe(withLatestFrom(search$, lastIndex$), map(function (_a) {
      var _b = __read(_a, 3),
          value = _b[0],
          searchInput = _b[1],
          lastIndexInput = _b[2];

      return value.lastIndexOf(searchInput, lastIndexInput);
    }));
  };
}

/**
 * Returns an Observable that emits a string from a source of character codes using String.fromCharCode
 *
 * @category Mapping
 *
 * @see The [[fromCharCode]] Observable can be used to generate a string source from character codes
 *
 * @example
 * Returns a string `RxJS` from a source array of character codes
 * ```ts
 * of([82, 120, 74, 83]).pipe(mapCharCode()).subscribe()
 * ```
 * Output: `RxJS`
 *
 *
 * @returns Observable that emits a string from source character codes
 */

function mapCharCode() {
  return function (source) {
    return source.pipe(map(function (values) {
      return typeof values === 'number' ? String.fromCharCode(values) : String.fromCharCode.apply(String, __spread(values));
    }));
  };
}

/**
 * Returns an Observable that emits a string from a source of character codes using String.fromCodePoint
 *
 * @category Mapping
 *
 * @see The [[fromCodePoint]] Observable can be used to generate a string source from character codes
 *
 * @example
 * Returns a string `RxJS` from a source array of code points
 * ```ts
 * of([9731, 9733, 9842]).pipe(mapCodePoint()).subscribe()
 * ```
 * Output: `☃★♲`
 *
 * @returns Observable that emits a string from source code points
 */

function mapCodePoint() {
  return function (source) {
    return source.pipe(map(function (values) {
      return typeof values === 'number' ? String.fromCodePoint(values) : String.fromCodePoint.apply(String, __spread(values));
    }));
  };
}

/**
 * Returns an Observable that emits a `RegExpMatchArray` where a source string returns a valid result using using
 * String.match. If no result is found, `null` is emitted.
 *
 * @category Query
 *
 * @param pattern A string or RegExp to match
 *
 * @example
 * Returns a single match for the string `RxJS` in the source string
 * ```ts
 * of('Hello RxJS Ninja, RxJS Rocks').pipe(match('RxJS')).subscribe();
 * ```
 * Output: `['RxJS', index: 6, input: 'Hello RxJS Ninja, RxJS Rocks', groups: undefined]`
 *
 * @example
 * Returns a match with a RegExp with global `/g` for `RxJS`
 * ```ts
 * of('Hello RxJS Ninja, RxJS Rocks').pipe(match(/RxJS/g)).subscribe();
 * ```
 * Output: `['RxJS', 'RxJS']`
 *
 * @returns Observable that emits a RegExpMatchArray
 */

function match(pattern) {
  var pattern$ = createOrReturnObservable(pattern);
  return function (source) {
    return source.pipe(withLatestFrom(pattern$), map(function (_a) {
      var _b = __read(_a, 2),
          value = _b[0],
          patternInput = _b[1];

      return value.match(patternInput);
    }));
  };
}

/**
 * Returns an Observable that emits an array of results from String.matchAll
 *
 * @category Query
 *
 * @remarks This operator converts the `IterableIterator<RegExpMatchArray>` to `RegExpMatchArray[]` to avoid dealing
 *   with iterators.
 *
 * @param pattern A RegExp regular expression to match in the string
 *
 * @example
 * Returns all matching groups in the source string
 * ```ts
 * of('RxJS Ninja,Hello Ninja').pipe(matchAll(/([A-Za-z\s]*)Ninja/g).subscribe();
 * ```
 *
 * Output:`["RxJS Ninja", index: 0, input: "RxJS Ninja,Hello Ninja"], ["Hello Ninja", index: 11, input: "RxJS
 *   Ninja,Hello Ninja"]`
 *
 * @returns Observable that emits an array of RegExpMatchArray
 */

function matchAll(pattern) {
  var pattern$ = createOrReturnObservable(pattern);
  return function (source) {
    return source.pipe(withLatestFrom(pattern$), map(function (_a) {
      var _b = __read(_a, 2),
          value = _b[0],
          patternInput = _b[1];

      return __spread(value.matchAll(patternInput));
    }));
  };
}

/**
 * Returns an Observable that emits a string made from a source unicode string using String.normalize
 *
 * @category Mapping
 *
 * @param form The Unicode Normalization Form to decode the string with
 *
 * @example
 * Returns a string from Unicode characters
 * ```ts
 * of('\u0041\u006d\u00e9\u006c\u0069\u0065').pipe(normalize()).subscribe();
 * ```
 * Output: `Amélie`
 *
 * @returns Observable that emits a string
 */

function normalize(form) {
  var form$ = createOrReturnObservable(form);
  return function (source) {
    return source.pipe(withLatestFrom(form$), map(function (_a) {
      var _b = __read(_a, 2),
          value = _b[0],
          formValue = _b[1];

      return value.normalize(formValue);
    }));
  };
}

/**
 * Returns an Observable that emits a string where the source string has been padded using String.padStart
 *
 * @category Modify
 *
 * @alias `padLeft`
 *
 * @param maxLength The maximum length to pad the string to
 * @param fillString Optional string to use as the string padding
 *
 * @example
 * Returns a string padded to a length of `12` with default fill string
 * ```ts
 * of('RxJS Ninja').pipe(padStart(12)).subscribe();
 * ```
 * Output: `'  RxJS Ninja'`
 *
 *
 * @example
 * Returns a string padded to a length of `12` with `.` fill string
 * ```ts
 * of('12345').pipe(padStart(12, '.')).subscribe();
 * ```
 * Output: `'..RxJS Ninja'`
 *
 * @returns Observable that emits a string that is padded to the passed length
 */

function padStart(maxLength, fillString) {
  var maxLength$ = createOrReturnObservable(maxLength);
  var fillString$ = createOrReturnObservable(fillString);
  return function (source) {
    return source.pipe(withLatestFrom(maxLength$, fillString$), map(function (_a) {
      var _b = __read(_a, 3),
          value = _b[0],
          maxLengthValue = _b[1],
          fillStringValue = _b[2];

      return value.padStart(maxLengthValue, fillStringValue);
    }));
  };
}
/**
 * Returns an Observable that emits a string where the source string has been padded using String.padEnd
 *
 * @category Modify
 *
 * @alias `padRight`
 *
 * @param maxLength The maximum length to pad the string to
 * @param fillString Optional string to use as the string padding
 *
 * @example
 * Returns a string padded to a length of `12` with default fill string
 * ```ts
 * of('RxJS Ninja').pipe(padEnd(12)).subscribe();
 * ```
 * Output: `'RxJS Ninja  '`
 *
 *
 * @example
 * Returns a string padded to a length of `12` with `.` fill string
 * ```ts
 * of('12345').pipe(padEnd(12, '.')).subscribe();
 * ```
 * Output: `'RxJS Ninja..'`
 *
 * @returns Observable that emits a string that is padded to the passed length
 */

function padEnd(maxLength, fillString) {
  var maxLength$ = createOrReturnObservable(maxLength);
  var fillString$ = createOrReturnObservable(fillString);
  return function (source) {
    return source.pipe(withLatestFrom(maxLength$, fillString$), map(function (_a) {
      var _b = __read(_a, 3),
          value = _b[0],
          maxLengthValue = _b[1],
          fillStringValue = _b[2];

      return value.padEnd(maxLengthValue, fillStringValue);
    }));
  };
}

/**
 * Returns an Observable that emits a string where the source string is repeated with String.repeat.
 *
 * @remarks If a separator is passed it uses an array and will join the result using the separator instead, as
 *   `String.repeat` does not support it
 *
 * @category Modify
 *
 * @param count The number of times to repeat the string
 * @param separator Optional separator for joining strings
 *
 * @example
 * Returns a string with the word `Ninja` repeated 5 times
 * ```ts
 * of('Ninja').pipe(repeat(5)).subscribe();
 * ```
 * Output: `NinjaNinjaNinjaNinjaNinja`
 *
 * @example
 * Returns a string with the word `Ninja` repeated 5 times with a `,` separator
 * ```ts
 * of('Ninja').pipe(repeat(5, ', ')).subscribe();
 * ```
 * Output: `Ninja, Ninja, Ninja, Ninja, Ninja`
 *
 * @returns Observable that emits a string of the source string repeated
 */

function repeat(count, separator) {
  var count$ = createOrReturnObservable(count);
  var separator$ = createOrReturnObservable(separator);
  return function (source) {
    return source.pipe(withLatestFrom(count$, separator$), map(function (_a) {
      var _b = __read(_a, 3),
          value = _b[0],
          countValue = _b[1],
          separatorValue = _b[2];

      return separatorValue ? new Array(countValue).fill(value).join(separatorValue) : value.repeat(countValue);
    }));
  };
}

/**
 * Returns an Observable that emits a string, replacing text in the source string with the replacement text if the
 * pattern is found using String.replace
 *
 * @category Modify
 *
 * @param pattern A string or RegExp to find in the Observable string to replace
 * @param replacement The replacement string
 *
 * @example
 * Return a string with first instance `Hero` replaced with `Ninja`
 * ```ts
 * of('RxJS Hero, Angular Hero').pipe(replace('Hero', 'Ninja')).subscribe();
 * ```
 * Output: `RxJS Ninja, Angular Hero`
 *
 * @example
 * Return a string where `Hero` is replaced with `Ninja` using a RegExp
 * ```ts
 * of('RxJS Hero, Angular Hero').pipe(replace(/(?!\w+\s)(\w+)/, 'Ninja')).subscribe();
 * ```
 * Output: `RxJS Ninja, Angular Hero`
 *
 * @returns Observable that emits a string
 */

function replace(pattern, replacement) {
  var pattern$ = createOrReturnObservable(pattern);
  var replacement$ = createOrReturnObservable(replacement);
  return function (source) {
    return source.pipe(withLatestFrom(pattern$, replacement$), map(function (_a) {
      var _b = __read(_a, 3),
          value = _b[0],
          patternValue = _b[1],
          replacementValue = _b[2];

      return value.replace(patternValue, replacementValue);
    }));
  };
}

/**
 * Returns an Observable that emits a string, replacing text in the source string with the replacement text if the
 * pattern is found using String.replaceAll
 *
 * @category Modify
 *
 * @param pattern A string or RegExp to find in the Observable string to replace
 * @param replacement The replacement string
 *
 * @example
 * Return a string with all instances of `Hero` replaced with `Ninja`
 * ```ts
 * of('RxJS Hero, Angular Hero').pipe(replaceAll('Hero', 'Ninja')).subscribe();
 * ```
 * Output: `RxJS Ninja, Angular Ninja`
 *
 * @example
 * Return a string where `Hero` is replaced with `Ninja` using a RegExp
 * ```ts
 * of('RxJS Hero, Angular Hero').pipe(replaceAll(/(?!\w+\s)(\w+)/g, 'Ninja')).subscribe();
 * ```
 * Output: `RxJS Ninja, Angular Ninja`
 *
 * @returns Observable that emits a string
 */

function replaceAll(pattern, replacement) {
  var pattern$ = createOrReturnObservable(pattern);
  var replacement$ = createOrReturnObservable(replacement);
  return function (source) {
    return source.pipe(withLatestFrom(pattern$, replacement$), map(function (_a) {
      var _b = __read(_a, 3),
          value = _b[0],
          patternValue = _b[1],
          replacementValue = _b[2];

      return value.replaceAll(patternValue, replacementValue);
    }));
  };
}

/**
 * Returns an Observable that emits a string where the source string is reversed.
 *
 * @category Modify
 *
 * @remarks This operator turns a string into an array and uses Array.reverse and Array.join to create
 * a new string
 *
 * @example
 * Returns a string that is reversed
 * ```ts
 * of('emordnilaP').pipe(reverse()).subscribe();
 * ```
 * Output: `'Palindrome'`
 *
 * @returns Observable that emits a string that is reversed from the source
 */

function reverse() {
  return function (source) {
    return source.pipe(map(function (value) {
      return __spread(value).reverse().join('');
    }));
  };
}

/**
 * Returns an Observable that emits a number that is the first index of where the value is found using String.search
 *
 * @category Query
 *
 * @param pattern A string or RegExp to match in the string
 *
 * @example
 * Return the first index of `Ninja`
 * ```ts
 * of('RxJS Ninja, Angular Ninja').pipe(search('Ninja')).subscribe();
 * ```
 * Output: `5`
 *
 * @example
 * Return the first index of the first found group from the RegExp
 * ```ts
 * of('RxJS Ninja, Angular Ninja').pipe(search(/(?!\w+\s)(\w+)/g)).subscribe();
 * ```
 * Output: `5`
 *
 * @returns Observable that emits an number that is the start index of the first found value
 */

function search(pattern) {
  var pattern$ = createOrReturnObservable(pattern);
  return function (source) {
    return source.pipe(withLatestFrom(pattern$), map(function (_a) {
      var _b = __read(_a, 2),
          value = _b[0],
          patternValue = _b[1];

      return value.search(patternValue);
    }));
  };
}

/**
 * Returns an Observable that emits a string that is a partial slice of the source string using String.slice
 *
 * @category Modify
 *
 * @remarks For creating substrings use [[substring]] as a better alternative to `slice`
 *
 * @param startIndex The start index for the substring
 * @param endIndex Optional end index for the length of substring, if not passed slice will use `str.length -1`
 *
 * @example
 * Return a string from index `0` to `4`
 * ```ts
 * of('RxJS Ninja').pipe(slice(0, 4)).subscribe();
 * ```
 * Output: `RxJS`
 *
 * @example
 * Return a string from index `5` to the end of the string
 * ```ts
 * of('RxJS Ninja').pipe(slice(5)).subscribe();
 * ```
 * Output: `Ninja`
 *
 * @returns Observable that emits a string that is a slice of the source string
 */

function slice(startIndex, endIndex) {
  var startIndex$ = createOrReturnObservable(startIndex);
  var endIndex$ = createOrReturnObservable(endIndex);
  return function (source) {
    return source.pipe(withLatestFrom(startIndex$, endIndex$), map(function (_a) {
      var _b = __read(_a, 3),
          value = _b[0],
          startIndexValue = _b[1],
          endIndexValue = _b[2];

      return value.slice(startIndexValue, endIndexValue);
    }));
  };
}

/**
 * Returns an Observable that emits an array from a source string split by the separator using String.split
 *
 * @category Convert
 *
 * @param separator The character to split the string at
 * @param limit Optional limit for the number of times to split
 *
 * @example
 * Returns an array of a string split on the `|` character
 * ```ts
 * of('Hello|World|RxJS|Ninja').pipe(split('|')).subscribe();
 * ```
 * Output: `['Hello', 'World', 'RxJS', 'Ninja']`
 *
 * @example
 * Returns an array of a string split on the `|` character, limiting to the first `2`
 * ```ts
 * of('Hello|World|RxJS|Ninja').pipe(split('|', 2)).subscribe();
 * ```
 * Output: `['Hello', 'World']`
 *
 * @returns Observable that emits an array of strings from the source string split on the separator
 */

function split(separator, limit) {
  if (separator === void 0) {
    separator = ' ';
  }

  var separator$ = createOrReturnObservable(separator);
  var limit$ = createOrReturnObservable(limit);
  return function (source) {
    return source.pipe(withLatestFrom(separator$, limit$), map(function (_a) {
      var _b = __read(_a, 3),
          value = _b[0],
          separatorValue = _b[1],
          limitValue = _b[2];

      return value.split(separatorValue, limitValue);
    }));
  };
}

/**
 * Returns an Observable that emits a boolean when the source string contains the input string at the start of the
 * source string using String.startsWith
 *
 * @category Query
 *
 * @param search The string to check the source string start with
 * @param startIndex Optional start index to being searching the string from
 *
 * @remarks
 * If you need to get the string value instead of boolean use [[filterStartsWith]]
 *
 * @example
 * Return a boolean if the source string starts with the passed input
 * ```ts
 * from(['RxJS', 'Ninja', 'Hero']).pipe(startsWith('R')).subscribe();
 * ```
 * Output: `true, false, false`
 *
 * @example
 * Return a boolean if the source string starts with the passed input from index `1`
 * ```ts
 * from(['RxJS', 'Ninja', 'Hero']).pipe(startsWith('x', 1)).subscribe();
 * ```
 * Output: `true, false, false`
 *
 * @returns Observable that emits a boolean if the source string start with the input string
 */

function startsWith(search, startIndex) {
  var search$ = createOrReturnObservable(search);
  var startIndex$ = createOrReturnObservable(startIndex);
  return function (source) {
    return source.pipe(withLatestFrom(search$, startIndex$), map(function (_a) {
      var _b = __read(_a, 3),
          value = _b[0],
          searchValue = _b[1],
          indexValue = _b[2];

      return value.startsWith(searchValue, indexValue);
    }));
  };
}

/**
 * Returns an Observable that emits a string that is a partial slice of the source string using String.substring
 *
 * @category Modify
 *
 * @param indexStart The index of the first character to include in the returned substring.
 * @param indexEnd Optional The index of the first character to exclude from the returned substring.
 *
 * @example
 * Return a string from index `0` to `4`
 * ```ts
 * of('RxJS Ninja').pipe(substring(0, 4)).subscribe();
 * ```
 * Output: `RxJS`
 *
 * @example
 * Return a string from index `5` to the end of the string
 * ```ts
 * of('RxJS Ninja').pipe(substring(5)).subscribe();
 * ```
 * Output: `Ninja`
 *
 * @returns Observable that emits a string that is a substring of the source string
 */

function substring(indexStart, indexEnd) {
  var indexStart$ = createOrReturnObservable(indexStart);
  var indexEnd$ = createOrReturnObservable(indexEnd);
  return function (source) {
    return source.pipe(withLatestFrom(indexStart$, indexEnd$), map(function (_a) {
      var _b = __read(_a, 3),
          value = _b[0],
          startValue = _b[1],
          endValue = _b[2];

      return value.substring(startValue, endValue);
    }));
  };
}

/**
 * Default words to exclude when using the [[titleize]] operator
 */

var NO_CAP_WORDS = ['a', 'and', 'but', 'is', 'or', 'over', 'the', 'to', 'too'];
/**
 * Return an Observable that emits a string where the source string is titleized (first letter of each word uppercase).
 * The operator uses String.toLocaleUpperCase so can support locale strings
 *
 * By default is uses the [[NO_CAP_WORDS]] to skip some words (unless they are the first word in the string).
 * To disable this pass an empty array, and to extend you can use the default with a spread operator to extend:
 * `[...NO_CAP_WORDS, 'foo', 'bar']`. When using a separator, if no change in default pass `undefined`
 *
 * It will also skip words already starting with a capital (e.g. `RxJS`)
 *
 * @category Modify
 *
 * @param noTitleWords A list of words to exclude from making a title word
 * @param separator Optional separator to use when joining each word
 * @param locales Locales for string formatting
 *
 * @example
 * Returns a string titilzed with default no caps words
 * ```ts
 * of('the RxJS ninja').pipe(titleize()).subscribe();
 * ```
 * Output: `The RxJS Ninja`
 *
 * @example
 * Returns a string titilzed with additional no cap words
 * ```ts
 * of('the RxJS ninja').pipe(titleize([...NO_CAP_WORDS, 'ninja'])).subscribe();
 * ```
 * Output: `The RxJS ninja`
 *
 * @example
 * Returns a string titilzed with a separator
 * ```ts
 * of('angular,RxJS,typescript,ninja').pipe(titleize(undefined, ',')).subscribe();
 * ```
 * Output: `Angular,RxJS,Typescript,Ninja`
 *
 * @example
 * Returns a string titilzed with a different locale
 * ```ts
 * of('ängulär,RxJS,typescript,ninjä').pipe(titleize(undefined, undefined, 'de-DE')).subscribe();
 * ```
 * Output: `Ängulär,RxJS,Typescript,Ninjä`
 *
 * @returns Observable that emits a titilzed string
 */

function titleize(noTitleWords, separator, locales) {
  if (noTitleWords === void 0) {
    noTitleWords = NO_CAP_WORDS;
  }

  if (separator === void 0) {
    separator = ' ';
  }

  var noTitleWords$ = createOrReturnObservable(noTitleWords);
  var locales$ = createOrReturnObservable(locales);
  return function (source) {
    return source.pipe(split(separator), // Here we use `split` passing the same value passed in to this operator
    withLatestFrom(noTitleWords$, locales$), map(function (_a) {
      var _b = __read(_a, 3),
          values = _b[0],
          noTitleWordsValue = _b[1],
          localesValues = _b[2];

      return __spread(values).map(function (word, index) {
        if (index && __spread(noTitleWordsValue).includes(word)) {
          return word;
        }

        return word.charCodeAt(0) >= 65 && word.charCodeAt(0) <= 90 ? word : "" + word.charAt(0).toLocaleUpperCase(localesValues) + word.slice(1);
      });
    }), join(separator));
  };
}

/**
 * Returns an Observable that emits a string where the source string is passed through String.toLocaleLowerCase
 *
 * @category Modify
 *
 * @param locales Optional locales to pass for string formatting
 *
 * @example
 * Returns an lower case string
 * ```ts
 * of('APPLE').pipe(toLowerCase()).subscribe();
 * ```
 * Output: `'apple'`
 *
 * @example
 *  * Returns an lower case string with locale
 * ```ts
 * of('ÄPFEL').pipe(toLowerCase('de-DE')).subscribe();
 * ```
 * Output: `'äpfel'`
 *
 * @returns Observable that emits a lower case string
 */

function toLowerCase(locales) {
  var locales$ = createOrReturnObservable(locales);
  return function (source) {
    return source.pipe(withLatestFrom(locales$), map(function (_a) {
      var _b = __read(_a, 2),
          value = _b[0],
          localesValue = _b[1];

      return value.toLocaleLowerCase(!localesValue || typeof localesValue === 'string' ? localesValue : __spread(localesValue));
    }));
  };
}

/**
 * Returns an Observable that emits a string where the source string is passed through String.toLocaleLowerCase
 *
 * @category Modify
 *
 * @param locales Optional locales to pass for string formatting
 *
 * @example
 * Returns an upper case string
 * ```ts
 * of('apple').pipe(toUpperCase()).subscribe();
 * ```
 * Output: `'APPLE'`
 *
 * @example
 * Returns an lower case string with locale
 * ```ts
 * of('äpfel').pipe(toUpperCase('de-DE')).subscribe();
 * ```
 * Output: `'ÄPFEL'`
 *
 * @returns Observable that emits a lower case string
 */

function toUpperCase(locales) {
  var locales$ = createOrReturnObservable(locales);
  return function (source) {
    return source.pipe(withLatestFrom(locales$), map(function (_a) {
      var _b = __read(_a, 2),
          value = _b[0],
          localesValue = _b[1];

      return value.toLocaleUpperCase(!localesValue || typeof localesValue === 'string' ? localesValue : __spread(localesValue));
    }));
  };
}

/**
 * Returns an Observable that emits a string where the source string has any white space at the start removed using
 * String.trimStart
 *
 * @category Modify
 *
 * @alias `trimLeft`
 *
 * @example
 * Returns the source string with any whitespace at the start removed
 * ```ts
 * of('  RxJS Ninja  ').pipe(trimStart()).subscribe();
 * ```
 * Output: `'RxJS Ninja  '`
 *
 * @returns Observable that emits a trimmed string
 */

function trimStart() {
  return function (source) {
    return source.pipe(map(function (value) {
      return value.trimStart();
    }));
  };
}
/**
 * Returns an Observable that emits a string where the source string has any white space at the end removed using
 * String.trimEnd
 *
 * @category Modify
 *
 * @alias `trimRight`
 *
 * @example
 * Returns the source string with any whitespace at the end removed
 * ```ts
 * of('  RxJS Ninja  ').pipe(trimStart()).subscribe();
 * ```
 * Output: `'  RxJS Ninja'`
 *
 * @returns Observable that emits a trimmed string
 */

function trimEnd() {
  return function (source) {
    return source.pipe(map(function (value) {
      return value.trimEnd();
    }));
  };
}
/**
 * Returns an Observable that emits a string where the source string has any white space at the ends removed using
 * String.trim
 *
 * @category Modify
 *
 * @example
 * Returns the source string with any whitespace at both ends removed
 * ```ts
 * of('  RxJS Ninja  ').pipe(trimStart()).subscribe();
 * ```
 * Output: `'RxJS Ninja'`
 *
 * @returns Observable that emits a trimmed string
 */

function trim() {
  return function (source) {
    return source.pipe(map(function (value) {
      return value.trim();
    }));
  };
}

/**
 * @packageDocumentation
 * @module String
 */

/**
 * The type of form to normalize a unicode string with String.normalize, default is 'NFC'
 */
var FormType;

(function (FormType) {
  /**
   * Canonical Decomposition, followed by Canonical Composition.
   */
  FormType["NFC"] = "NFC";
  /**
   * Canonical Decomposition.
   */

  FormType["NFD"] = "NFD";
  /**
   * Compatibility Decomposition, followed by Canonical Composition.
   */

  FormType["NFKC"] = "NFKC";
  /**
   * Compatibility Decomposition.
   */

  FormType["NFKD"] = "NFKD";
})(FormType || (FormType = {}));

export { FormType, NO_CAP_WORDS, charAt, charCodeAt, codePointAt, concat, endsWith, filterEndsWith, filterIncludes, filterStartsWith, fromCharCode, fromCodePoint, fromString, fromUnicode, includes, indexOf, join, lastIndexOf, mapCharCode, mapCodePoint, match, matchAll, normalize, padEnd, padStart as padLeft, padEnd as padRight, padStart, repeat, replace, replaceAll, reverse, search, slice, split, startsWith, substring, titleize, toLowerCase, toUpperCase, trim, trimEnd, trimStart as trimLeft, trimEnd as trimRight, trimStart };
